/*
 * File: utils.c
 * Date: 2011-11-17
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 * MACOS port by Aria Stewart <aredridel@nbtsc.org>
 */

#include	<ctype.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include        <time.h>
#include	"rho.h"
#include	"utils.h"
#include	"bind.h"
#include	"display.h"
#include	"exec.h"
#include	"fileio.h"
#include	"messages.h"

#if MACOS
/* nop */
#else
extern int asprintf(char **, char *, ...);
#endif                          /* MACOS */

/*++

Get a keystroke command sequence from the keyboard. Process all applicable
prefix keys.

++*/
int get_keystrokes()
{
    int c;                      /* fetched keystroke */
    KEYTAB *key;                /* ptr to a key entry */

    /* get initial character */
    c = mlgetkey();
    key = exist_binding(c);

    /* resolve META and CTLX prefixes */
    if (key) {
        if (key->k_fp == meta_prefix) {
            c = mlgetkey();
            c = toupper(c) | (c & ~255);        /* Force to upper */
            c |= META;
        } else if (key->k_fp == ctlx) {
            c = mlgetkey();
            c = toupper(c) | (c & ~255);        /* Force to upper */
            c |= CTLX;
        }
    }

    /* return it */
    return (c);
}



/*++

Returns index of string "key" in array from fn "tval" of "tlength" elements.

++*/
int index_of_str(char *key, char *(*tval) (), int tlength)
{
    int l, u;                   /* lower and upper limits of binary search */
    int i;                      /* current search index */
    int cresult;                /* result of comparison */

    /* set current search limit as entire list */
    l = 0;
    u = tlength - 1;

    /* get the midpoint! */
    while (u >= l) {
        i = (l + u) >> 1;

        /* do the comparison */
        cresult = strcmp(key, (*tval) (i));
        if (cresult == 0)
            return (i);
        if (cresult < 0)
            u = i - 1;
        else
            l = i + 1;
    }
    return (-1);
}



/*++

Global helper converting null-pointer to empty string.

++*/
char *fixnull(s)                /* Don't return NULL pointers! */
char *s;

{
    if (s == NULL)
        return ("");
    else
        return (s);
}



/*++

Converts string val to int val.  Was here for portability.  Changing to
libc.strtol would be a lot of work:

    long int strtol (const char *restrict string, char **restrict tailptr, int base)

++*/
int str2int(st)
char *st;

{
    int result;                 /* resulting number */
    int sign;                   /* sign of resulting number */
    char c;                     /* current char being examined */

    result = 0;
    sign = 1;

    /* skip preceding whitespace */
    while (*st == ' ' || *st == '\t')
        ++st;

    /* check for sign */
    if (*st == '-') {
        sign = -1;
        ++st;
    }
    if (*st == '+')
        ++st;

    /* scan digits, build value */
    while ((c = *st++))
        if (c >= '0' && c <= '9')
            result = result * 10 + c - '0';
        else
            break;

    return (result * sign);
}



/*++

Self-explanatory. Was here for portability. Now an asprintf wrapper.

++*/
char *int2str(i)
int i;                          /* integer to translate to a string */

{

    char *sp;
    asprintf(&sp, "%d", i);
    return (sp);
}



/*++

Convert a string to a numeric boolean value
Local helper

++*/
int str2bool(val)
char *val;                      /* value to check for str2bool */

{
    /* check for logical values */
    if (val[0] == 'F')
        return (FALSE);
    if (val[0] == 'T')
        return (TRUE);

    /* check for numeric truth (!= 0) */
    return ((str2int(val) != 0));
}



/*++

Convert a numeric boolean value to a string const.
Local helper

++*/
char *bool2str(val)
int val;                        /* value to translate */

{
    if (val)
        return ((char *) truem);
    else
        return ((char *) falsem);
}



/*++

toupper wrapper for pointers
Global helper

++*/
void toupperp(cp)
unsigned char *cp;              /* ptr to character to uppercase */

{
    if (islower(*cp))
        *cp = (unsigned char) toupper(*cp);
}



/*++

tolower wrapper for pointers
Local helper

++*/
void tolowerp(cp)
unsigned char *cp;              /* ptr to character to lowercase */

{
    if (isupper(*cp))
        *cp = (unsigned char) tolower(*cp);
}



/*++

UFUNC to make string uppercase

++*/
char *mkupper(str)              /* make a string upper case */
char *str;                      /* string to upper case */

{
    char *sp;

    sp = str;
    while (*sp)
        toupperp((unsigned char *) sp++);
    return (str);
}



/*++

UFUNC to make string lowercase
Also local helper

++*/
char *mklower(str)              /* make a string lower case */
char *str;                      /* string to lower case */

{
    char *sp;

    sp = str;
    while (*sp)
        tolowerp((unsigned char *) sp++);
    return (str);
}


/*++

Expanded character to character collapses the CTRL and SPEC flags back
into an ascii code.

++*/
int ectoc(c)
int c;

{
    if (c == (CTRL | ' '))
        c = 0;
    if (c & CTRL)
        c = c ^ (CTRL | 0x40);
    if (c & SPEC)
        c = c & 255;
    return (c);
}

/*++

Character to extended character pulls out the CTRL and SPEC prefixes.

++*/
int ctoec(c)
int c;

{
    if (c == 0)
        c = CTRL | ' ';
    else if ((c >= 0x00 && c <= 0x1F) || c == 0x7F)
        c = CTRL | (c ^ 0x40);
    return (c);
}



/*++

Get a key from the terminal driver, resolve any keyboard macro action.

++*/
int tgetc()
{
    int c;                      /* fetched character */

    /* if we are playing a keyboard macro back, */
    if (kbdmode == PLAY) {

        /* if there is some left... */
        if (kbdptr < kbdend)
            return ((int) *kbdptr++);

        /* at the end of last repitition? */
        if (--kbdrep < 1) {
            kbdmode = STOP;
            /* force a screen update after all is done */
            update(FALSE);
        } else {

            /* reset the macro to the begining for the next rep */
            kbdptr = &kbdm[0];
            return ((int) *kbdptr++);
        }
    }

    /* if no pending character */
    if (cpending == FALSE) {

        /* fetch a character from the terminal driver */
        c = TTgetc();

    } else {

        c = charpending;
        cpending = FALSE;
    }

    /* record it for $lastkey */
    lastkey = c;

    /* save it if we need to */
    if (kbdmode == RECORD) {
        *kbdptr++ = c;
        kbdend = kbdptr;

        /* don't overrun the buffer */
        if (kbdptr == &kbdm[NKBDM - 1]) {
            kbdmode = STOP;
            TTbeep();
        }
    }

    /* and finally give the char back */
    return (c);
}



/*++

Function that does NOTHING

++*/
int nop(f, n)
int n, f;

{

    f = n = 0;                  /* splint */
    return f;
}


/*++

Dummy function for binding to universal-argument.
Does even less than nop()

++*/
int universal_argument()
{
    return (0);
}


/*++

Beep the beeper and do a general aborting of stuff.

++*/
int abort_command(f, n)
int f, n;                       /* prefix flag and argument */

{

    f = n = 0;                  /* splint */

    TTbeep();
    kbdmode = STOP;
    mlwrite(TEXT8);
/*		"[Aborted]" */
    return (ABORT);
}



/*++

Tell the user that this command is illegal while we are in VIEW
(read-only) mode.

++*/

int err_readonly()
{
    TTbeep();
    mlwrite(TEXT109);
/*		"[Key illegal in VIEW mode]" */
    return (FALSE);
}



/*++

Set META prefixing pending

++*/
int meta_prefix(f, n)
int f, n;                       /* prefix flag and argument */

{
    prefix |= META;
    prenum = n;
    predef = f;
    return (TRUE);
}



/*++

Set ^X prefixing pending

++*/
int ctlx(f, n)
int f, n;                       /* prefix flag and argument */

{
    prefix |= CTLX;
    prenum = n;
    predef = f;
    return (TRUE);
}



/*++

Toggle overwrite/insert with insert key

++*/
int toggle_overwrite(void)
{
    if (curbp->b_mode & MDOVER) {
        curbp->b_mode &= ~(1 << 5);
    } else {
        curbp->b_mode |= (1 << 5);
        curbp->b_mode &= ~MDREPL;
    }
    update_modeline();
    return (TRUE);
}



/*++

Get index of fn_string from names[]. Helper for get_fn_name.

++*/
static char *fn_string_index(index)
int index;

{
    return (names[index].n_name);
}



/*++

In rho_name.h we bind "name" to fn_name().
Takes a ptr to string and gets the "name" associated with it.
Global helper fn.
	
++*/
int (*get_fn_name(fname)) ()
char *fname;                    /* name to attempt to match */

{
    int nval;

    if ((nval = index_of_str(fname, fn_string_index, numfunc)) == -1)
        return (NULL);
    else
        return (names[nval].n_func);
}



/*++

Execute what is bound to a key_cmd (k_code in a KEYTAB.)
Global helper fn.

++*/
int execute_key(key, f, n)
KEYTAB *key;                    /* key to execute */
int f, n;                       /* agruments to C function */

{
    int status;                 /* error return */

    if (key->k_type == BINDFNC)
        return ((*(key->k_fp)) (f, n));
    if (key->k_type == BINDBUF) {
        while (n--) {
            status = do_buffer(key->k_buf);
            if (status != TRUE)
                return (status);
        }
    }
    return (TRUE);
}



/*++

Global helper function for finding a file to open.  Opens an absolute path
or fails.  Relative paths search $HOME, then $PATH, then bind.path_name

++*/
char *search_paths(fname, hflag)
char *fname;                    /* base file name to search for */
int hflag;                      /* Look in the HOME environment variable first? */

{
    char *home;                 /* path to home directory */
    char *path;                 /* environmental PATH variable */
    char *sp;                   /* pointer into path spec */
    int i;                      /* index */
    static char fspec[LENPATH]; /* full path spec to search */

    /* if we have an absolute path, check only there */
    sp = fname;
    while (*sp) {
        if (*sp == DIRSEPCHAR) {
            if (fexist(fname)) {
                return (fname);
            } else
                return (NULL);
        }
        ++sp;
    }


    if (hflag) {
        home = getenv("HOME");
        if (home != NULL) {
            /* build home dir file spec */
            strcpy(fspec, home);
            strcat(fspec, DIRSEPSTR);
            strcat(fspec, fname);

            /* and try it out */
            if (fexist(fspec)) {
                return (fspec);
            }
        }
    }

    /* current directory now overides everything except HOME var */
    if (fexist(fname)) {
        return (fname);
    }

    /* get the PATH variable */
    path = getenv("PATH");
    if (path != NULL)
        while (*path) {

            /* build next possible file spec */
            sp = fspec;
            while (*path && (*path != PATHCHR))
                *sp++ = *path++;

            /* add a terminating dir separator if we need it */
            if ((sp != fspec) && (*(sp - 1) != DIRSEPCHAR))
                *sp++ = DIRSEPCHAR;
            *sp = 0;
            strcat(fspec, fname);

            /* and try it out */
            if (fexist(fspec)) {
                return (fspec);
            }

            if (*path == PATHCHR)
                ++path;
        }

    /* look it up via the old table method */
    for (i = 0; i < npnames; i++) {
        strcpy(fspec, search_path[i]);
        strcat(fspec, fname);

        /* and try it out */
        if (fexist(fspec)) {
            return (fspec);
        }
    }

    return (NULL);              /* no such luck */
}




/*++

	A key binding consists of one or more prefix functions followed by
	a keystroke.  Allowable prefixes must be in the following order:

	^X	preceeding control-X
	A-	similtaneous ALT key (on PCs mainly)
	S-	shifted function key
	M-	Preceding META key
	FN	function key
	^	control key

	Meta and ^X prefix of lower case letters are converted to upper
	case.  Real control characters are automatically converted to
	the ^A form.

++*/
int keystr2keycmd(keyname)
char *keyname;                  /* name of key to translate to Command key form */

{
    int c;                      /* key sequence to return */

    /* parse it up */
    c = 0;

    /* Do ^X prefix */
    if (*keyname == '^' && *(keyname + 1) == 'X') {
        if (*(keyname + 2) != 0) {      /* Key is not bare ^X */
            c |= CTLX;
            keyname += 2;
        }
    }

    /* and the SHIFTED prefix */
    if (*keyname == 'S' && *(keyname + 1) == '-') {
        c |= SHFT;
        keyname += 2;
    }

    /* then the META prefix */
    if (*keyname == 'M' && *(keyname + 1) == '-') {
        c |= META;
        keyname += 2;
    }

    /* next the function prefix */
    if (*keyname == 'F' && *(keyname + 1) == 'N') {
        c |= SPEC;
        keyname += 2;
    }

    /* a control char?  (NOT Always upper case anymore) */
    if (*keyname == '^' && *(keyname + 1) != 0) {
        c |= CTRL;
        ++keyname;
        if (*keyname == '@')
            *keyname = ' ';
    }

    /* A literal control character? */
    if (*keyname < 32) {
        c |= CTRL;
        *keyname += '@';
    }

    /* make sure we are not lower case if used with ^X or M- */
    if (!(c & (SPEC | SHFT)))   /* If not a special key */
        if (c & (CTLX | META))  /* If is a prefix */
            toupperp((unsigned char *) keyname);        /* Then ensure toupperp */

    /* the final sequence... */
    c |= *keyname;
    return (c);
}



/*++

In rho_name.h we bind "name" to fn_name().
Takes a ptr to KEYTAB entry and gets the "name" associated with it.
Global helper fn.
	
++*/
char *get_name(key)
KEYTAB *key;                    /* key binding to return a name of */

{
    int (*func) ();             /* ptr to the requested function */
    NBIND *nptr;                /* pointer into the name binding table */
    BUFFER *bp;                 /* ptr to buffer to test */
    BUFFER *kbuf;               /* ptr to requested buffer */

    /* if this isn't a valid key, it has no name */
    if (key == NULL)
        return (NULL);

    /* skim through the binding table, looking for a match */
    if (key->k_type == BINDFNC) {
        func = key->k_fp;
        nptr = &names[0];
        while (nptr->n_func != NULL) {
            if (nptr->n_func == func)
                return (nptr->n_name);
            ++nptr;
        }
        return (NULL);
    }

    /* skim through the buffer list looking for a match */
    if (key->k_type == BINDBUF) {
        kbuf = key->k_buf;
        bp = bheadp;
        while (bp) {
            if (bp == kbuf)
                return (bp->b_name);
            bp = bp->b_bufp;
        }
        return (NULL);
    }
    return (NULL);
}



/*++

Get time of day

++*/
char *getctime()
{
    time_t buf;
    char *sp, *cp;

    /* Get system time */
    (void) time(&buf);

    /* Pass system time to converter */
    sp = ctime(&buf);

    /* Eat newline character */
    for (cp = sp; *cp; cp++)
        if (*cp == '\n') {
            *cp = '\0';
            break;
        }
    return (sp);
}
